/*
	modplayer en punto fijo para la Gameboy Advance
	Copyright (C) 2005  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <string.h>
#include "gba.h"
#include "palette.h"
#include "font_8x8.h"



#define BG0_ENABLE      0x100
#define BG1_ENABLE      0x200
#define BG2_ENABLE      0x400
#define BG3_ENABLE      0x800
#define OBJ_ENABLE      0x1000
#define WIN0_ENABLE     0x2000
#define WIN1_ENABLE     0x4000
#define WINOBJ_ENABLE   0x8000

#define BACK_BUFFER     0x10

#define MODE0           0x0
#define MODE1           0x1
#define MODE2           0x2
#define MODE3           0x3
#define MODE4           0x4
#define MODE5           0x5

#define  VIDEO_BUFFER_0  ((u16 *) 0x6000000)
#define  VIDEO_BUFFER_1  ((u16 *) 0x600A000)
#define  BG_PALETTE_MEM  ((u16 *) 0x5000000)

#define  TOP_LEFT      0
#define  TOP_RIGHT     239
#define  BOTTOM_LEFT   38060
#define  BOTTOM_RIGHT  38299

#define RGB16(r,g,b)  ((r) + ((g) << 5) + ((b) << 10))

#define  ERASE_SCREEN(scr)  memset(scr, 0, 240 * 160)




#define  SAMPLE_RATE  25600
#define  SAMPLES_PER_TICK  (SAMPLE_RATE / 50)   // we choose 25600 Hz because we obtain a 2-power sized buffer (512)


#define  BUFFER_SIZE    (SAMPLES_PER_TICK * 2)
#define  BUFFER_SIZE_2  (BUFFER_SIZE / 2)
#define  BUFFER_SIZE_4  (BUFFER_SIZE / 4)
#define  BUFFER_SIZE_8  (BUFFER_SIZE / 8)

u32 buffer[BUFFER_SIZE_4];
u32 *buffer_middle = &(buffer[BUFFER_SIZE_8]);
int sub_buffer = 0;


#define  NUM_NOTES  60

#define  NTSC_FACTOR  7159090.5
#define  PAL_FACTOR   7093789.2

#define  NTSC_FACTOR_FP  ((unsigned int) 1832727168)    // NTSC_FACTOR * 256
#define  PAL_FACTOR_FP   ((unsigned int) 1816010035)    // PAL_FACTOR * 256

// the mod file is linked directly into the executable
extern unsigned char battle_mod[];

unsigned short volume_table[65] = {
    0,     4,     8,    12,    16,    20,    24,    28,    32,    36,    40,    44,    48,    52,    56,    60,
   64,    68,    72,    76,    80,    84,    88,    92,    96,   100,   104,   108,   112,   116,   120,   124,
  128,   131,   136,   141,   144,   147,   152,   156,   160,   163,   167,   171,   175,   179,   183,   188,
  192,   196,   201,   203,   208,   213,   215,   220,   223,   228,   231,   236,   239,   244,   247,   253,
  256
};

unsigned short period_table[16][NUM_NOTES] = {
	{1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016, 960 , 906,
	 856 , 808 , 762 , 720 , 678 , 640 , 604 , 570 , 538 , 508 , 480 , 453,
	 428 , 404 , 381 , 360 , 339 , 320 , 302 , 285 , 269 , 254 , 240 , 226,
	 214 , 202 , 190 , 180 , 170 , 160 , 151 , 143 , 135 , 127 , 120 , 113,
	 107 , 101 , 95  , 90  , 85  , 80  , 75  , 71  , 67  , 63  , 60  , 56 },
    {1700, 1604, 1514, 1430, 1348, 1274, 1202, 1134, 1070, 1010, 954 , 900,
     850 , 802 , 757 , 715 , 674 , 637 , 601 , 567 , 535 , 505 , 477 , 450,
     425 , 401 , 379 , 357 , 337 , 318 , 300 , 284 , 268 , 253 , 239 , 225,
     213 , 201 , 189 , 179 , 169 , 159 , 150 , 142 , 134 , 126 , 119 , 113,
     106 , 100 , 94  , 89  , 84  , 79  , 75  , 71  , 67  , 63  , 59  , 56 },
    {1688, 1592, 1504, 1418, 1340, 1264, 1194, 1126, 1064, 1004, 948 , 894,
     844 , 796 , 752 , 709 , 670 , 632 , 597 , 563 , 532 , 502 , 474 , 447,
     422 , 398 , 376 , 355 , 335 , 316 , 298 , 282 , 266 , 251 , 237 , 224,
     211 , 199 , 188 , 177 , 167 , 158 , 149 , 141 , 133 , 125 , 118 , 112,
     105 , 99  , 94  , 88  , 83  , 79  , 74  , 70  , 66  , 62  , 59  , 56 },
    {1676, 1582, 1492, 1408, 1330, 1256, 1184, 1118, 1056, 996 , 940 , 888,
     838 , 791 , 746 , 704 , 665 , 628 , 592 , 559 , 528 , 498 , 470 , 444,
     419 , 395 , 373 , 352 , 332 , 314 , 296 , 280 , 264 , 249 , 235 , 222,
     209 , 198 , 187 , 176 , 166 , 157 , 148 , 140 , 132 , 125 , 118 , 111,
     104 , 99  , 93  , 88  , 83  , 78  , 74  , 70  , 66  , 62  , 59  , 55 },
    {1664, 1570, 1482, 1398, 1320, 1246, 1176, 1110, 1048, 990 , 934 , 882,
     832 , 785 , 741 , 699 , 660 , 623 , 588 , 555 , 524 , 495 , 467 , 441,
     416 , 392 , 370 , 350 , 330 , 312 , 294 , 278 , 262 , 247 , 233 , 220,
     208 , 196 , 185 , 175 , 165 , 156 , 147 , 139 , 131 , 124 , 117 , 110,
     104 , 98  , 92  , 87  , 82  , 78  , 73  , 69  , 65  , 62  , 58  , 55 },
    {1652, 1558, 1472, 1388, 1310, 1238, 1168, 1102, 1040, 982 , 926 , 874,
     826 , 779 , 736 , 694 , 655 , 619 , 584 , 551 , 520 , 491 , 463 , 437,
     413 , 390 , 368 , 347 , 328 , 309 , 292 , 276 , 260 , 245 , 232 , 219,
     206 , 195 , 184 , 174 , 164 , 155 , 146 , 138 , 130 , 123 , 116 , 109,
     103 , 97  , 92  , 87  , 82  , 77  , 73  , 69  , 65  , 61  , 58  , 54 },
    {1640, 1548, 1460, 1378, 1302, 1228, 1160, 1094, 1032, 974 , 920 , 868,
     820 , 774 , 730 , 689 , 651 , 614 , 580 , 547 , 516 , 487 , 460 , 434,
     410 , 387 , 365 , 345 , 325 , 307 , 290 , 274 , 258 , 244 , 230 , 217,
     205 , 193 , 183 , 172 , 163 , 154 , 145 , 137 , 129 , 122 , 115 , 109,
     102 , 96  , 91  , 86  , 81  , 77  , 72  , 68  , 64  , 61  , 57  , 54 },
    {1628, 1536, 1450, 1368, 1292, 1220, 1150, 1086, 1026, 968 , 914 , 862,
     814 , 768 , 725 , 684 , 646 , 610 , 575 , 543 , 513 , 484 , 457 , 431,
     407 , 384 , 363 , 342 , 323 , 305 , 288 , 272 , 256 , 242 , 228 , 216,
     204 , 192 , 181 , 171 , 161 , 152 , 144 , 136 , 128 , 121 , 114 , 108,
     102 , 96  , 90  , 85  , 80  , 76  , 72  , 68  , 64  , 60  , 57  , 54 },
    {1814, 1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016, 960,
     907 , 856 , 808 , 762 , 720 , 678 , 640 , 604 , 570 , 538 , 508 , 480,
     453 , 428 , 404 , 381 , 360 , 339 , 320 , 302 , 285 , 269 , 254 , 240,
     226 , 214 , 202 , 190 , 180 , 170 , 160 , 151 , 143 , 135 , 127 , 120,
     113 , 107 , 101 , 95  , 90  , 85  , 80  , 75  , 71  , 67  , 63  , 60 },
    {1800, 1700, 1604, 1514, 1430, 1350, 1272, 1202, 1134, 1070, 1010, 954,
     900 , 850 , 802 , 757 , 715 , 675 , 636 , 601 , 567 , 535 , 505 , 477,
     450 , 425 , 401 , 379 , 357 , 337 , 318 , 300 , 284 , 268 , 253 , 238,
     225 , 212 , 200 , 189 , 179 , 169 , 159 , 150 , 142 , 134 , 126 , 119,
     112 , 106 , 100 , 94  , 89  , 84  , 79  , 75  , 71  , 67  , 63  , 59 },
    {1788, 1688, 1592, 1504, 1418, 1340, 1264, 1194, 1126, 1064, 1004, 948,
     894 , 844 , 796 , 752 , 709 , 670 , 632 , 597 , 563 , 532 , 502 , 474,
     447 , 422 , 398 , 376 , 355 , 335 , 316 , 298 , 282 , 266 , 251 , 237,
     223 , 211 , 199 , 188 , 177 , 167 , 158 , 149 , 141 , 133 , 125 , 118,
     111 , 105 , 99  , 94  , 88  , 83  , 79  , 74  , 70  , 66  , 62  , 59 },
    {1774, 1676, 1582, 1492, 1408, 1330, 1256, 1184, 1118, 1056, 996 , 940,
     887 , 838 , 791 , 746 , 704 , 665 , 628 , 592 , 559 , 528 , 498 , 470,
     444 , 419 , 395 , 373 , 352 , 332 , 314 , 296 , 280 , 264 , 249 , 235,
     222 , 209 , 198 , 187 , 176 , 166 , 157 , 148 , 140 , 132 , 125 , 118,
     111 , 104 , 99  , 93  , 88  , 83  , 78  , 74  , 70  , 66  , 62  , 59 },
    {1762, 1664, 1570, 1482, 1398, 1320, 1246, 1176, 1110, 1048, 988 , 934,
     881 , 832 , 785 , 741 , 699 , 660 , 623 , 588 , 555 , 524 , 494 , 467,
     441 , 416 , 392 , 370 , 350 , 330 , 312 , 294 , 278 , 262 , 247 , 233,
     220 , 208 , 196 , 185 , 175 , 165 , 156 , 147 , 139 , 131 , 123 , 117,
     110 , 104 , 98  , 92  , 87  , 82  , 78  , 73  , 69  , 65  , 61  , 58 },
    {1750, 1652, 1558, 1472, 1388, 1310, 1238, 1168, 1102, 1040, 982 , 926,
     875 , 826 , 779 , 736 , 694 , 655 , 619 , 584 , 551 , 520 , 491 , 463,
     437 , 413 , 390 , 368 , 347 , 328 , 309 , 292 , 276 , 260 , 245 , 232,
     219 , 206 , 195 , 184 , 174 , 164 , 155 , 146 , 138 , 130 , 123 , 116,
     109 , 103 , 97  , 92  , 87  , 82  , 77  , 73  , 69  , 65  , 61  , 58 },
    {1736, 1640, 1548, 1460, 1378, 1302, 1228, 1160, 1094, 1032, 974 , 920,
     868 , 820 , 774 , 730 , 689 , 651 , 614 , 580 , 547 , 516 , 487 , 460,
     434 , 410 , 387 , 365 , 345 , 325 , 307 , 290 , 274 , 258 , 244 , 230,
     217 , 205 , 193 , 183 , 172 , 163 , 154 , 145 , 137 , 129 , 122 , 115,
     108 , 102 , 96  , 91  , 86  , 81  , 77  , 72  , 68  , 64  , 61  , 57 },
    {1724, 1628, 1536, 1450, 1368, 1292, 1220, 1150, 1086, 1026, 968 , 914,
     862 , 814 , 768 , 725 , 684 , 646 , 610 , 575 , 543 , 513 , 484 , 457,
     431 , 407 , 384 , 363 , 342 , 323 , 305 , 288 , 272 , 256 , 242 , 228,
     216 , 203 , 192 , 181 , 171 , 161 , 152 , 144 , 136 , 128 , 121 , 114,
     108 , 101 , 96  , 90  , 85  , 80  , 76  , 72  , 68  , 64  , 60  , 57 }
};

unsigned int inc_table_25600_hz[16][NUM_NOTES] = {
	{ 20,   21,   23,   24,   26,   27,   29,   31,   32,   34,   36,   39, 
	  41,   43,   46,   49,   52,   55,   58,   62,   65,   69,   73,   78, 
	  82,   87,   93,   98,  104,  110,  117,  124,  131,  139,  147,  156, 
	 165,  175,  186,  197,  208,  221,  234,  248,  262,  279,  295,  313, 
	 331,  351,  373,  394,  417,  443,  472,  499,  529,  562,  591,  633},
	{ 20,   22,   23,   24,   26,   27,   29,   31,   33,   35,   37,   39, 
	  41,   44,   46,   49,   52,   55,   59,   62,   66,   70,   74,   78, 
	  83,   88,   93,   99,  105,  111,  118,  124,  132,  140,  148,  157, 
	 166,  176,  187,  198,  209,  223,  236,  249,  264,  281,  298,  313, 
	 334,  354,  377,  398,  422,  448,  472,  499,  529,  562,  601,  633},
	{ 21,   22,   23,   25,   26,   28,   29,   31,   33,   35,   37,   39, 
	  42,   44,   47,   50,   52,   56,   59,   62,   66,   70,   74,   79, 
	  84,   89,   94,   99,  105,  112,  119,  125,  133,  141,  149,  158, 
	 168,  178,  188,  200,  212,  224,  238,  251,  266,  283,  300,  316, 
	 337,  358,  377,  403,  427,  448,  479,  506,  537,  572,  601,  633},
	{ 21,   22,   23,   25,   26,   28,   29,   31,   33,   35,   37,   39, 
	  42,   44,   47,   50,   53,   56,   59,   63,   67,   71,   75,   79, 
	  84,   89,   95,  100,  106,  112,  119,  126,  134,  142,  150,  159, 
	 169,  179,  189,  201,  213,  225,  239,  253,  268,  283,  300,  319, 
	 341,  358,  381,  403,  427,  454,  479,  506,  537,  572,  601,  644},
	{ 21,   22,   23,   25,   26,   28,   30,   31,   33,   35,   37,   40, 
	  42,   45,   47,   50,   53,   56,   60,   63,   67,   71,   75,   80, 
	  85,   90,   95,  101,  107,  113,  120,  127,  135,  143,  152,  161, 
	 170,  180,  191,  202,  214,  227,  241,  255,  270,  286,  303,  322, 
	 341,  361,  385,  407,  432,  454,  485,  514,  545,  572,  611,  644},
	{ 21,   22,   24,   25,   27,   28,   30,   32,   34,   36,   38,   40, 
	  42,   45,   48,   51,   54,   57,   60,   64,   68,   72,   76,   81, 
	  85,   90,   96,  102,  108,  114,  121,  128,  136,  144,  152,  161, 
	 172,  181,  192,  203,  216,  228,  242,  257,  272,  288,  305,  325, 
	 344,  365,  385,  407,  432,  460,  485,  514,  545,  581,  611,  656},
	{ 21,   22,   24,   25,   27,   28,   30,   32,   34,   36,   38,   40, 
	  43,   45,   48,   51,   54,   57,   61,   64,   68,   72,   77,   81, 
	  86,   91,   97,  102,  109,  115,  122,  129,  137,  145,  154,  163, 
	 173,  183,  193,  206,  217,  230,  244,  258,  274,  290,  308,  325, 
	 347,  369,  389,  412,  437,  460,  492,  521,  554,  581,  622,  656},
	{ 21,   23,   24,   25,   27,   29,   30,   32,   34,   36,   38,   41, 
	  43,   46,   48,   51,   54,   58,   61,   65,   69,   73,   77,   82, 
	  87,   92,   97,  103,  109,  116,  123,  130,  138,  146,  155,  164, 
	 173,  184,  195,  207,  220,  233,  246,  260,  277,  293,  311,  328, 
	 347,  369,  394,  417,  443,  466,  492,  521,  554,  591,  622,  656},
	{ 19,   20,   21,   23,   24,   26,   27,   29,   31,   32,   34,   36, 
	  39,   41,   43,   46,   49,   52,   55,   58,   62,   65,   69,   73, 
	  78,   82,   87,   93,   98,  104,  110,  117,  124,  131,  139,  147, 
	 156,  165,  175,  186,  197,  208,  221,  234,  248,  262,  279,  295, 
	 313,  331,  351,  373,  394,  417,  443,  472,  499,  529,  562,  591},
	{ 19,   20,   22,   23,   24,   26,   27,   29,   31,   33,   35,   37, 
	  39,   41,   44,   46,   49,   52,   55,   59,   62,   66,   70,   74, 
	  78,   83,   88,   93,   99,  105,  111,  118,  124,  132,  140,  149, 
	 157,  167,  177,  187,  198,  209,  223,  236,  249,  264,  281,  298, 
	 316,  334,  354,  377,  398,  422,  448,  472,  499,  529,  562,  601},
	{ 19,   21,   22,   23,   25,   26,   28,   29,   31,   33,   35,   37, 
	  39,   42,   44,   47,   50,   52,   56,   59,   62,   66,   70,   74, 
	  79,   84,   89,   94,   99,  105,  112,  119,  125,  133,  141,  149, 
	 159,  168,  178,  188,  200,  212,  224,  238,  251,  266,  283,  300, 
	 319,  337,  358,  377,  403,  427,  448,  479,  506,  537,  572,  601},
	{ 19,   21,   22,   23,   25,   26,   28,   29,   31,   33,   35,   37, 
	  39,   42,   44,   47,   50,   53,   56,   59,   63,   67,   71,   75, 
	  79,   84,   89,   95,  100,  106,  112,  119,  126,  134,  142,  150, 
	 159,  169,  179,  189,  201,  213,  225,  239,  253,  268,  283,  300, 
	 319,  341,  358,  381,  403,  427,  454,  479,  506,  537,  572,  601},
	{ 20,   21,   22,   23,   25,   26,   28,   30,   31,   33,   35,   37, 
	  40,   42,   45,   47,   50,   53,   56,   60,   63,   67,   71,   75, 
	  80,   85,   90,   95,  101,  107,  113,  120,  127,  135,  143,  152, 
	 161,  170,  180,  191,  202,  214,  227,  241,  255,  270,  288,  303, 
	 322,  341,  361,  385,  407,  432,  454,  485,  514,  545,  581,  611},
	{ 20,   21,   22,   24,   25,   27,   28,   30,   32,   34,   36,   38, 
	  40,   42,   45,   48,   51,   54,   57,   60,   64,   68,   72,   76, 
	  81,   85,   90,   96,  102,  108,  114,  121,  128,  136,  144,  152, 
	 161,  172,  181,  192,  203,  216,  228,  242,  257,  272,  288,  305, 
	 325,  344,  365,  385,  407,  432,  460,  485,  514,  545,  581,  611},
	{ 20,   21,   22,   24,   25,   27,   28,   30,   32,   34,   36,   38, 
	  40,   43,   45,   48,   51,   54,   57,   61,   64,   68,   72,   77, 
	  81,   86,   91,   97,  102,  109,  115,  122,  129,  137,  145,  154, 
	 163,  173,  183,  193,  206,  217,  230,  244,  258,  274,  290,  308, 
	 328,  347,  369,  389,  412,  437,  460,  492,  521,  554,  581,  622},
	{ 20,   21,   23,   24,   25,   27,   29,   30,   32,   34,   36,   38, 
	  41,   43,   46,   48,   51,   54,   58,   61,   65,   69,   73,   77, 
	  82,   87,   92,   97,  103,  109,  116,  123,  130,  138,  146,  155, 
	 164,  174,  184,  195,  207,  220,  233,  246,  260,  277,  293,  311, 
	 328,  351,  369,  394,  417,  443,  466,  492,  521,  554,  591,  622}
};


typedef struct {
	signed char *data;
	int length;
	unsigned char fine_tune;
	unsigned char volume;
	int repeat_offset;
	int repeat_length;
} mod_sample_header;

typedef struct {
	signed char *sample_data;
	unsigned char volume;
	unsigned int inc, t;
	int sample_size;
	int sample_repeat_offset;
	int sample_repeat_length;
	unsigned char trig;
} sample_status;

typedef union {
	unsigned int word;
	unsigned char byte[4];
} mod_row;

typedef struct {
	mod_sample_header sample_header[31];
	unsigned char order_size;
	unsigned char *order;                  // 128 entries
	unsigned char num_channels;
	unsigned char num_patterns;
	mod_row *pattern;
	short int current_order;
	short int current_row;
	short int current_velocity;
	short int tick_count;
	sample_status channel[8];
} mod_file;

mod_file song;

void InterruptProcess(void) __attribute__ ((section(".iwram")));
void put_string(char *s, int x, int y, u16 *video_buffer, int erase);

#define  REVERSE_ENDIAN(v)  ((((v) << 8) & 0xFF00) | (((v) >> 8) & 0x00FF))

#define  MOD_SAMPLE_HEADER_OFFSET  20
#define  MOD_SAMPLE_HEADER_SIZE    30
#define  MOD_POST_SAMPLE_HEADER_OFFSET  (20 + (MOD_SAMPLE_HEADER_SIZE * 31))


void InterruptProcess(void) {
	/* se acabó el sample */
	if (sub_buffer == 0) {
		/* Terminó de reproducirse la primera mitad del buffer */
		REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
		REG_DMA1SAD = (unsigned long) buffer_middle; /* nuevo buffer origen para el DMA1 (la segunda mitad del buffer) */
		REG_DMA1CNT_H = 0xb600; /* reactivamos el DMA1 + FIFO + 32 bits + con repetición */
		sub_buffer = 1;
		/*
		screen_buffer[TOP_LEFT] = RGB16(31, 0, 0);
		screen_buffer[TOP_RIGHT] = RGB16(0, 0, 0);
		*/
	}
	else {
		/* terminó de reproducirse la segunda mitad del buffer */
		REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
		REG_DMA1SAD = (unsigned long) buffer; /* nuevo buffer origen para el DMA1 (la primera mitad del buffer) */
		REG_DMA1CNT_H = 0xb600; /* reactivamos el DMA1 + FIFO + 32 bits + con repetición */
		sub_buffer = 0;
		/*
		screen_buffer[TOP_LEFT] = RGB16(0, 0, 0);
		screen_buffer[TOP_RIGHT] = RGB16(31, 0, 0);
		*/
	}
	/* borramos siempre el flag de interrupción */
	REG_IF |= REG_IF;
}


void dma_play_sound(void) {
	REG_TM0CNT_H = 0; /* deshabilitamos el timer 0*/
	/* reproducimos un sonido mono a 16KHz en modo DMA */
	/* utilizamos el timer 0 como generador del período de muestreo */
	/* y el timer 1 en cascada para contar los samples que se van enviando a la FIFO del chip de sonido */
	REG_SOUNDCNT_L = 0;
	REG_SOUNDCNT_H = 0x0b0f; /* Direct Sound A&B + reiniciar la FIFO + timer0 + volumen máximo a los canales L y R */
	REG_SOUNDCNT_X = 0x0080; /* activamos el chip de sonido */

	REG_DMA1SAD = (unsigned long) buffer; /* buffer origen para el DMA1 */
	REG_DMA1DAD = 0x040000a0; /* buffer destino = dirección del puerto FIFO del chip de sonido */
	REG_DMA1CNT_H = 0xb600; /* activamos el DMA1 + FIFO + 32 bits + con repetición */

	REG_TM1CNT_L = 0xffff - BUFFER_SIZE_2; /* 0xffff - número de muestras (la primera mitad del buffer) */
	REG_TM1CNT_H = 0xC4; /* habilitamos el timer 1 + irq y en cascada con el timer 0 */

	REG_IE = 0x10; /* habilitar interrupción al desbordarse el timer 1 */
	REG_IME = 1; /* habilitar las interrupciones */

	/* La fórmula para obtener el valor del timer 0: 0xFFFF - round(frecCPU / frecMuestreo) siendo frecCPU = 16MHz */
	//REG_TM0CNT_L = 0xFBE8; /* 16kHz */
	REG_TM0CNT_L = 0xFD70; /* 25.6kHz */
	REG_TM0CNT_H = 0x0080; /* habilitar timer 0 a la frecuencia de la CPU */
	/* indicamos que estamos reproduciendo la primera mitad del buffer */
	sub_buffer = 0;
}


void dma_stop_sound(void) {
	REG_TM0CNT_H = 0; /* deshabilitamos el timer 0 */
	REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
}


void start_playing_mod(unsigned char *buffer, mod_file *tune) {
	int i, j, aux_offset;
	unsigned short aux;
	unsigned char aux_buffer[80];
	unsigned char max;

	memcpy(aux_buffer, buffer, 20);
	aux_buffer[20] = 0;
	put_string(aux_buffer, 0, 0, VIDEO_BUFFER_0, 0);
	for (i = 0; i < 31; i++) {
		// read sample size
		memcpy(&aux, &(buffer[MOD_SAMPLE_HEADER_OFFSET + (i * MOD_SAMPLE_HEADER_SIZE) + 22]), 2);
#if  BYTE_ORDER == LITTLE_ENDIAN
		tune->sample_header[i].length = ((int) REVERSE_ENDIAN(aux)) << 1;
#else
		tune->sample_header[i].length = ((int) aux) << 1;
#endif
		// read fine tune
		tune->sample_header[i].fine_tune = buffer[MOD_SAMPLE_HEADER_OFFSET + (i * MOD_SAMPLE_HEADER_SIZE) + 24];
		// read volume
		tune->sample_header[i].volume = buffer[MOD_SAMPLE_HEADER_OFFSET + (i * MOD_SAMPLE_HEADER_SIZE) + 25];
		// read repeat offset
		memcpy(&aux, &(buffer[MOD_SAMPLE_HEADER_OFFSET + (i * MOD_SAMPLE_HEADER_SIZE) + 26]), 2);
#if  BYTE_ORDER == LITTLE_ENDIAN
		tune->sample_header[i].repeat_offset = ((int) REVERSE_ENDIAN(aux)) << 1;
#else
		tune->sample_header[i].repeat_offset = ((int) aux) << 1;
#endif
		// read repeat length
		memcpy(&aux, &(buffer[MOD_SAMPLE_HEADER_OFFSET + (i * MOD_SAMPLE_HEADER_SIZE) + 28]), 2);
#if  BYTE_ORDER == LITTLE_ENDIAN
		tune->sample_header[i].repeat_length = ((int) REVERSE_ENDIAN(aux)) << 1;
#else
		tune->sample_header[i].repeat_length = ((int) aux) << 1;
#endif
	}
	// read number of song positions
	tune->order_size = buffer[MOD_POST_SAMPLE_HEADER_OFFSET];
	// trash
	// read(f, &aux, 1);
	// read order table
	tune->order = &(buffer[MOD_POST_SAMPLE_HEADER_OFFSET + 2]);
	max = 0;
	for (i = 0; i < tune->order_size; i++) {
		if (tune->order[i] > max)
			max = tune->order[i];
	}
	tune->num_patterns = max + 1;
	// read signature
	memcpy(aux_buffer, &(buffer[MOD_POST_SAMPLE_HEADER_OFFSET + 130]), 4);
	aux_buffer[4] = 0;
	put_string(aux_buffer, 0, 8, VIDEO_BUFFER_0, 0);
	if ((strcmp(aux_buffer, "6CHN") == 0) || (strcmp(aux_buffer, "FLT6") == 0))
		tune->num_channels = 6;
	else if ((strcmp(aux_buffer, "8CHN") == 0) || (strcmp(aux_buffer, "FLT8") == 0))
		tune->num_channels = 8;
	else
		tune->num_channels = 4;
	// patterns
	tune->pattern = (mod_row *) &(buffer[MOD_POST_SAMPLE_HEADER_OFFSET + 134]);
	/*
	 * uncomment this to dump the pattern data to screen at startup
	 *
	for (i = 0; i < tune->num_patterns; i++) {
		printf("pattern %d:\n", i);
		for (j = 0; j < 64; j++) {
			for (channel = 0; channel < tune->num_channels; channel++) {
				unsigned char row_buffer[4];
				unsigned char sample;
				unsigned short period;
				unsigned short effect;
				unsigned int row_offset = (i * 64 * tune->num_channels) + (j * tune->num_channels) + channel;
				memcpy(row_buffer, &(tune->pattern[row_offset]), 4);
				sample = (row_buffer[2] >> 4) | (row_buffer[0] & 0xF0);
				period = (((unsigned short) row_buffer[1]) | (((unsigned short) row_buffer[0]) << 8)) & 0x0FFF;
				effect = (((unsigned short) row_buffer[3]) | (((unsigned short) row_buffer[2]) << 8)) & 0x0FFF;
				printf(" %03d %05d %3x", sample, period, effect);
			}
			printf("\n");
		}
		printf("\n");
	}
	*/
	// init sample data
	aux_offset = MOD_POST_SAMPLE_HEADER_OFFSET + 134 + (tune->num_channels * tune->num_patterns * 64 * 4);
	j = 0;
	for (i = 0; i < 31; i++) {
		tune->sample_header[i].data = &(buffer[aux_offset + j]);
		j += tune->sample_header[i].length;
	}
	tune->current_order = 0;
	tune->current_row = 0;
	tune->current_velocity = 6;
	tune->tick_count = 0;
	for (i = 0; i < tune->num_channels; i++) {
		tune->channel[i].sample_data = NULL;
		tune->channel[i].trig = 0;
	}
}


unsigned int search_inc(unsigned short period, unsigned char fine_tune) {
	unsigned short *table = period_table[fine_tune & 0x0F];
	unsigned int *inc_table = inc_table_25600_hz[fine_tune & 0x0F];
	int left = 0;
	int right = NUM_NOTES - 1;
	while (left <= right) {
		int center = (left + right) >> 1;
		if (period == table[center])
			return inc_table[center];
		else if (period < table[center])
			left = center + 1;
		else
			right = center - 1;
	}
	return inc_table[left];
}


void play_sample_start(sample_status *status, mod_sample_header *header, unsigned short period) {
	// unsigned short real_period = search_period(period, header->fine_tune);
	status->sample_data = header->data;
	status->sample_size = header->length;
	status->sample_repeat_offset = header->repeat_offset;
	status->sample_repeat_length = header->repeat_length;
	// Hay que cambiar esta instrucción para que la frecuencia de muestreo sea recogida de una tabla precalculada
	// status->inc = (PAL_FACTOR_FP / (real_period << 1)) / SAMPLE_RATE;
	status->inc = search_inc(period, header->fine_tune);
	status->t = 0;
	status->trig = 1;
	status->volume = header->volume;
}


void play_sample_add(sample_status *status, signed short int *output) {
	int i;
	int sample_real_size = (status->sample_repeat_length > 1) ? (status->sample_repeat_offset + status->sample_repeat_length) : status->sample_size;
	if (!status->trig)
		return;
	for (i = 0; i < SAMPLES_PER_TICK; i++) {
		signed short int v = status->sample_data[status->t >> 8];
		output[i] += (v * volume_table[status->volume]) >> 8;
		status->t += status->inc;
		if ((status->t >> 8) >= sample_real_size) {
			if (status->sample_repeat_length > 1)
				status->t = ((unsigned int) status->sample_repeat_offset) << 8;
			else {
				status->trig = 0;
				break;
			}
		}
	}
}


void tick_playing_mod(mod_file *song, signed char *buffer) {
	static signed short int mix_buffer[SAMPLES_PER_TICK];
	unsigned char c;
	int i;
	mod_row row_buffer;

	if (song->tick_count == 0) {
		// printf("order=%02d  pattern=%02d  row=%02d\n", song->current_order, song->order[song->current_order], song->current_row);
		int current_pattern = song->order[song->current_order];
		unsigned int row_offset = (current_pattern * 64 * song->num_channels) + ((int)song->current_row * song->num_channels);
		for (c = 0; c < song->num_channels; c++) {
			unsigned char sample;
			unsigned short period;
			unsigned short effect;
			unsigned char effect_type;
			unsigned char effect_value;
			row_buffer = song->pattern[row_offset + c];
			sample = (row_buffer.byte[2] >> 4) | (row_buffer.byte[0] & 0xF0);
			period = (((unsigned short) row_buffer.byte[1]) | (((unsigned short) row_buffer.byte[0]) << 8)) & 0x0FFF;
			effect = (((unsigned short) row_buffer.byte[3]) | (((unsigned short) row_buffer.byte[2]) << 8)) & 0x0FFF;
			effect_type = effect >> 8;
			effect_value = effect & 0x00FF;
			if (period > 0)
				play_sample_start(&(song->channel[c]), &(song->sample_header[sample - 1]), period);
			// check effects
			if (effect_type == 0x0F) {
				if ((effect_value >= 1) && (effect_value <= 31))
					song->current_velocity = effect_value;
			}
			else if (effect_type == 0x0C) {
				song->channel[c].volume = (effect_value > 64) ? 64 : effect_value;
			}
		}
		song->tick_count = song->current_velocity;
		if (song->current_row == 63) {
			song->current_row = 0;
			if (song->current_order == (song->order_size - 1)) {
				// loop song
				song->current_order = 0;
			}
			else
				song->current_order++;
		}
		else
			song->current_row++;
	}
	else
		song->tick_count--;
	for (i = 0; i < SAMPLES_PER_TICK; i++)
		mix_buffer[i] = 0;
	// bzero(mix_buffer, SAMPLES_PER_TICK * sizeof(signed short int));
	for (c = 0; c < song->num_channels; c++)
		play_sample_add(&(song->channel[c]), mix_buffer);
	for (i = 0; i < SAMPLES_PER_TICK; i++) {
		// buffer[i] = mix_buffer[i] / song->num_channels;
		register short int v = mix_buffer[i] >> 2;
		buffer[i] = (v > 127) ? 127 : ((v < -128) ? -128 : v);
	}
}


void put_character(char c, int x, int y, u16 *video_buffer, int erase) {
	int font_y = (((int) c) - 32) / 16;
	int font_x = (((int) c) - 32) % 16;
	int i, j;
	int font_offset = (font_y * 64 * 8) + (font_x * 4);
	int offset = ((y * 240) + x) >> 1;
	for (i = 0; i < 8; i++, offset += 120, font_offset += 64) {
		if (erase) {
			for (j = 0; j < 4; j++)
				video_buffer[offset + j] = 0x0000;
		}
		for (j = 0; j < 4; j++) {
			register u16 d = fontData[font_offset + j];
			register u16 prev = video_buffer[offset + j];
			register u16 to_write = 0;
			if ((d & 0x00FF) != fontTransparentIndex)
				to_write |= (d & 0x00FF);
			else
				to_write |= (prev & 0x00FF);
			if ((d >> 8) != fontTransparentIndex)
				to_write |= (d & 0xFF00);
			else
				to_write |= (prev & 0xFF00);
			video_buffer[offset + j] = to_write;
		}
	}
}


void put_string(char *s, int x, int y, u16 *video_buffer, int erase) {
	int i = 0;
	while (s[i] != 0) {
		put_character(s[i], x, y, video_buffer, erase);
		x += 8;
		i++;
	}
}



int AgbMain(void) {
	u16 i, last_written = 0;

	REG_DISPCNT = MODE4 | BG2_ENABLE;
	for (i = 0; i < 256; i++)
		BG_PALETTE_MEM[i] = my_palette[i];
	start_playing_mod(battle_mod, &song);
	dma_play_sound();
	while (1) {
		if (sub_buffer == 0) {
			if (last_written == 0) {
				//put_string("0", 0, 16, VIDEO_BUFFER_0, 1);
				// si se está reproduciendo la primera mitad del buffer, rellenamos la segunda
				tick_playing_mod(&song, (signed char *) buffer_middle);
				last_written = 1;
			}
		}
		else {
			if (last_written == 1) {
				//put_string("1", 0, 16, VIDEO_BUFFER_0, 1);
				// si se está reproduciendo la segunda mitad del buffer, rellenamos la primera
				tick_playing_mod(&song, (signed char *) buffer);
				last_written = 0;
			}
		}
	}
	return 0;
}
